/*
 * Copyright (C) 2004-2005 by David J. Hardy.  All rights reserved.
 *
 * nheptic3.c
 */

#include <stdio.h>
#include "mgrid/split.h"
/* #define DEBUG_WATCH */
#include "debug/debug.h"

#define NELEMS(a)  (sizeof(a)/sizeof((a)[0]))

/* prototypes of internal routines */
static void done(Mgrid *mg);
static int force(Mgrid *mg, MgridSystem *sys);
static int atoms_to_grid(Mgrid *mg, const MgridSystem *sys);
static int grid_to_atoms(Mgrid *mg, MgridSystem *sys);
static int restrict_grid(Mgrid *mg, int32 level);
static int prolong_grid(Mgrid *mg, int32 level);
static int direct_sum(Mgrid *mg, int32 level);


int mgrid_nheptic3_setup(Mgrid *mg, const MgridSystem *sys)
{
#if 0
  const double *q = sys->charge;
  const int32 nlevels = mg->param.nlevels;
  const int32 natoms = mg->param.natoms;
  const int32 split = mg->param.split;
  double *gd;
  double r2, s, t, gs, gt;
  int32 i, j, k, n, na, nb;
  int32 gdim, index;
  const double a = mg->param.cutoff;
  const double h = mg->param.spacing;
  const double h_1 = 1./h;
#endif

  /* assign virtual methods */
  mg->long_force = force;
  mg->long_done = done;

  return mgrid_setup_longrange(mg, sys);

#if 0
  /* set inverse of grid spacing */
  mg->inv_spacing = h_1;

  /* set location of multi-grids origin with lattice label (0,0,0) */
  mg->origin.x = mg->param.center.x - 0.5 * mg->param.length;
  mg->origin.y = mg->param.center.y - 0.5 * mg->param.length;
  mg->origin.z = mg->param.center.z - 0.5 * mg->param.length;

#ifdef DEBUG_WATCH
  printf("origin = (%g, %g, %g)\n", mg->origin.x, mg->origin.y, mg->origin.z);
#endif

  /* allocate arrays of charge and potential lattices */
  mg->qgrid = (MgridLattice *) malloc(nlevels * sizeof(MgridLattice));
  mg->egrid = (MgridLattice *) malloc(nlevels * sizeof(MgridLattice));
  if (mg->qgrid == NULL || mg->egrid == NULL) return MGRID_FAIL;

  /* setup charge and potential lattices for each level */
  for (k = 0;  k < nlevels;  k++) {
    if (mgrid_lattice_init(&(mg->qgrid[k]))
        || mgrid_lattice_init(&(mg->egrid[k]))) {
      return MGRID_FAIL;
    }
  }
  na = -3;
  nb = mg->param.nspacings + 3;
  for (k = 0;  k < nlevels;  k++) {
    if (nb - na + 1 <= 14 && k < nlevels - 1) {
      return MGRID_FAIL;
    }
    if (mgrid_lattice_setup(&(mg->qgrid[k]), sizeof(double),
          na, nb, na, nb, na, nb)
        || mgrid_lattice_setup(&(mg->egrid[k]), sizeof(double),
          na, nb, na, nb, na, nb)) {
      return MGRID_FAIL;
    }
#ifdef DEBUG_WATCH
    printf("level:  k = %d   na = %d   nb = %d\n", k, na, nb);
#endif
    na = -((-na+1)/2) - 3;
    nb = (nb+1)/2 + 3;
  }

  /* allocate and compute scalings for each direct sum on each level */
  mg->scaling = (double *) malloc(nlevels * sizeof(double));
  if (mg->scaling == NULL) return MGRID_FAIL;
  mg->scaling[0] = 1.0;
  for (k = 1;  k < nlevels;  k++) {
    mg->scaling[k] = 0.5 * mg->scaling[k-1];
  }

  /* allocate space for direct sum weights */
  n = (int32) (2.0 * a * h_1);
  mg->gdsum_radius = n;
  if (mgrid_lattice_setup(&(mg->gdsum), sizeof(double), -n, n, -n, n, -n, n)) {
    return MGRID_FAIL;
  }

  /* compute direct sum weights */
  if (mgrid_lattice_zero(&(mg->gdsum))) {
    return MGRID_FAIL;
  }
  gdim = mg->gdsum.ni;
  ASSERT(gdim == 2 * n + 1);
  gd = (double *)(mg->gdsum.data);
  for (k = -n;  k <= n;  k++) {
    for (j = -n;  j <= n;  j++) {
      for (i = -n;  i <= n;  i++) {
        r2 = (i*i +  j*j + k*k) * h*h;
        if (r2 >= 4*a*a) continue;
        s = r2 / (a*a);
        t = r2 / (4*a*a);
        ASSERT(t <= s);
        index = (k * gdim + j) * gdim + i;
        ASSERT(&gd[index] == mgrid_lattice_elem(&(mg->gdsum), i, j, k));
        if (s <= 1) gamma(&gs, s, split);
        else gs = 1/sqrt(s);
        if (t < 1) {
          gamma(&gt, t, split);
          gd[index] = gs/a - gt/(2*a);
        }
        else {
          /* then gt=1/sqrt(t), which implies that gs/a - gt(2*a) == 0 */
          gd[index] = 0;
        }
      }
    }
  }

  /* allocate space for direct sum weights on last level */
  n = mg->qgrid[nlevels-1].ni - 1;
  mg->glast_radius = n;
  if (mgrid_lattice_setup((&mg->glast), sizeof(double), -n, n, -n, n, -n, n)) {
    return MGRID_FAIL;
  }

  /* compute direct sum weights on last level */
  if (mgrid_lattice_zero(&(mg->glast))) {
    return MGRID_FAIL;
  }
  gdim = mg->glast.ni;
  ASSERT(gdim == 2 * n + 1);
  gd = (double *)(mg->glast.data);
  for (k = -n;  k <= n;  k++) {
    for (j = -n;  j <= n;  j++) {
      for (i = -n;  i <= n;  i++) {
        r2 = (i*i + j*j + k*k) * h*h;
        s = r2 / (a*a);
        index = (k * gdim + j) * gdim + i;
        ASSERT(&gd[index] == mgrid_lattice_elem(&(mg->glast), i, j, k));
        /* unlike previous, this sum is not truncated */
        if (s <= 1) gamma(&gs, s, split);
        else gs = 1/sqrt(s);
        gd[index] = gs/a;
      }
    }
  }

  /* compute g_a(0) */
  s = 0;
  gamma(&gs, s, split);
  mg->g_zero = gs/a;

  /* compute self potential */
  s = 0;
  for (n = 0;  n < natoms;  n++) {
    s += q[n] * q[n];
  }
  mg->u_self = 0.5 * mg->g_zero * s;

  return 0;
#endif
}


void done(Mgrid *mg)
{
  int32 n;

  mgrid_lattice_done(&(mg->gdsum));
  mgrid_lattice_done(&(mg->glast));
  for (n = 0;  n < mg->param.nlevels;  n++) {
    mgrid_lattice_done(&(mg->qgrid[n]));
    mgrid_lattice_done(&(mg->egrid[n]));
  }
  free(mg->qgrid);
  free(mg->egrid);
  free(mg->scaling);
}


int force(Mgrid *mg, MgridSystem *sys)
{
  int32 n;

  /* first charge grid must be zeroed */
  if (mgrid_lattice_zero(&(mg->qgrid[0]))) return MGRID_FAIL;

  /* computation is "inverse V-cycle" */
  if (atoms_to_grid(mg, sys)) return MGRID_FAIL;
  INT(mg->param.nlevels);
  for (n = 0;  n < mg->param.nlevels - 1;  n++) {
    if (direct_sum(mg, n)) return MGRID_FAIL;
    if (restrict_grid(mg, n)) return MGRID_FAIL;
  }
  INT(n);
  if (direct_sum(mg, n)) return MGRID_FAIL;
  for (--n;  n >= 0;  --n) {
    if (prolong_grid(mg, n)) return MGRID_FAIL;
  }
  if (grid_to_atoms(mg, sys)) return MGRID_FAIL;

  return 0;
}


int atoms_to_grid(Mgrid *mg, const MgridSystem *sys)
{
  double xphi[8], yphi[8], zphi[8];    /* phi grid func along x, y, z */
  double dx_h, dy_h, dz_h;             /* distance between atom and gridlo */
  double t;                            /* param to phi function */
  double ck, cjk;
  const double h_1 = mg->inv_spacing;  /* inverse grid spacing */
  const MD_Dvec origin = mg->origin;   /* lowest corner of domain */
  const double *q = sys->charge;       /* atom charges */
  const MD_Dvec *p = sys->pos;         /* atom positions */

  double *qh = (double *)(mg->qgrid[0].data);
      /* treat charge grid as 3D-lattice, index (i,j,k) */

  const int32 ndim = mg->qgrid[0].ni;  /* number points in each dimension */
  const int32 natoms = mg->param.natoms;  /* number of atoms */
  int32 n, i, j, k, ilo, jlo, klo;
  int32 index, koff, jkoff;

  for (n = 0;  n < natoms;  n++) {

    /* distance between atom and corner measured in grid points */
    dx_h = (p[n].x - origin.x) * h_1;
    dy_h = (p[n].y - origin.y) * h_1;
    dz_h = (p[n].z - origin.z) * h_1;

    /* find smallest numbered grid point in stencil */
    ilo = ((int32) dx_h) - 3;
    jlo = ((int32) dy_h) - 3;
    klo = ((int32) dz_h) - 3;

    /* find t for x dimension and compute xphi */
    t = dx_h - (double) ilo;
    xphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    t--;
    xphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    t--;
    xphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    t--;
    xphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    t--;
    xphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    t--;
    xphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    t--;
    xphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    t--;
    xphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));

    /* find t for y dimension and compute yphi */
    t = dy_h - (double) jlo;
    yphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    t--;
    yphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    t--;
    yphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    t--;
    yphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    t--;
    yphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    t--;
    yphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    t--;
    yphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    t--;
    yphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));

    /* find t for z dimension and compute zphi */
    t = dz_h - (double) klo;
    zphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    t--;
    zphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    t--;
    zphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    t--;
    zphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    t--;
    zphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    t--;
    zphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    t--;
    zphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    t--;
    zphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));

    /* determine charge on 64=4*4*4 qh grid point stencil */
    for (k = 0;  k < 8;  k++) {
      koff = (k + klo) * ndim;
      ck = zphi[k] * q[n];
      for (j = 0;  j < 8;  j++) {
        jkoff = (koff + (j + jlo)) * ndim;
        cjk = yphi[j] * ck;
        for (i = 0;  i < 8;  i++) {
          index = jkoff + (i + ilo);
          ASSERT(&qh[index]
              == mgrid_lattice_elem(&(mg->qgrid[0]), i+ilo, j+jlo, k+klo));
          qh[index] += xphi[i] * cjk;
        }
      }
    }

  } /* end loop over n */

#ifdef DEBUG_WATCH
  mgrid_lattice_print(&(mg->qgrid[0]));
#endif

  return 0;
}


int grid_to_atoms(Mgrid *mg, MgridSystem *sys)
{
  double xphi[8], yphi[8], zphi[8];     /* phi grid func along x, y, z */
  double dxphi[8], dyphi[8], dzphi[8];  /* dphi grid func along x, y, z */
  double dx_h, dy_h, dz_h;              /* distance between atom and gridlo */
  double t;                             /* param to phi and dphi functions */
  MD_Dvec c, fv;
  double u;
  const double u_self = mg->u_self;     /* self potential */
  const double h_1 = mg->inv_spacing;   /* inverse grid spacing */
  const MD_Dvec origin = mg->origin;    /* lowest corner of domain */
  const double *q = sys->charge;        /* atom charges */
  const MD_Dvec *p = sys->pos;          /* atom positions */
  MD_Dvec *f = (sys->f_long ? sys->f_long : sys->f_elec);
      /* accumulate contributions to atom force in f_long if allocated */

  double *eh = (double *)(mg->egrid[0].data);
      /* treat potential grid as 3D-lattice, index (i,j,k) */

  const double *evh = (double *)(mg->egrid[0].databuffer);
  const double *qvh = (double *)(mg->qgrid[0].databuffer);
      /* treat potential and charge grids as 1D arrays, index n */

  const int32 nelems = mg->egrid[0].nelems;  /* total number of grid points */
  const int32 ndim = mg->egrid[0].ni;    /* number points in each dimension */
  const int32 natoms = mg->param.natoms;     /* number of atoms */

  int32 n, i, j, k, ilo, jlo, klo;
  int32 index, koff, jkoff;

  /* compute force */
  for (n = 0;  n < natoms;  n++) {

    /* distance between atom and corner measured in grid points */
    dx_h = (p[n].x - origin.x) * h_1;
    dy_h = (p[n].y - origin.y) * h_1;
    dz_h = (p[n].z - origin.z) * h_1;

    /* find largest numbered grid point in stencil */
    ilo = ((int32) dx_h) - 3;
    jlo = ((int32) dy_h) - 3;
    klo = ((int32) dz_h) - 3;

    /* find t for x dimension, compute xphi and dxphi */ 
    t = dx_h - (double) ilo;
    xphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    dxphi[0] = ((-7456./5) + t*((117572./45) + t*(-1899 + t*((26383./36)
              + t*((-22807./144) + t*((727./40) + t*(-623./720))))))) * h_1;
    t--;
    xphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    dxphi[1] = ((25949./20) + t*((-117131./36) + t*((6741./2)
            + t*((-66437./36) + t*((81109./144) + t*((-727./8)
                  + t*(4361./720))))))) * h_1;
    t--;
    xphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    dxphi[2] = ((-8617./60) + t*((12873./20) + t*((-2373./2) + t*((4557./4)
              + t*((-9583./16) + t*((6543./40) + t*(-4361./240))))))) * h_1;
    t--;
    xphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    dxphi[3] = (t*((-49./18) + t*t*((-959./36) + t*((12845./144)
              + t*((-727./8) + t*(4361./144)))))) * h_1;
    t--;
    xphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    dxphi[4] = (t*((-49./18) + t*t*((-959./36) + t*((-12845./144)
              + t*((-727./8) + t*(-4361./144)))))) * h_1;
    t--;
    xphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    dxphi[5] = ((8617./60) + t*((12873./20) + t*((2373./2) + t*((4557./4)
              + t*((9583./16) + t*((6543./40) + t*(4361./240))))))) * h_1;
    t--;
    xphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    dxphi[6] = ((-25949./20) + t*((-117131./36) + t*((-6741./2)
            + t*((-66437./36) + t*((-81109./144) + t*((-727./8)
                  + t*(-4361./720))))))) * h_1;
    t--;
    xphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));
    dxphi[7] = ((7456./5) + t*((117572./45) + t*(1899 + t*((26383./36)
              + t*((22807./144) + t*((727./40) + t*(623./720))))))) * h_1;

    /* find t for y dimension, compute yphi and dyphi */
    t = dy_h - (double) jlo;
    yphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    dyphi[0] = ((-7456./5) + t*((117572./45) + t*(-1899 + t*((26383./36)
              + t*((-22807./144) + t*((727./40) + t*(-623./720))))))) * h_1;
    t--;
    yphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    dyphi[1] = ((25949./20) + t*((-117131./36) + t*((6741./2)
            + t*((-66437./36) + t*((81109./144) + t*((-727./8)
                  + t*(4361./720))))))) * h_1;
    t--;
    yphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    dyphi[2] = ((-8617./60) + t*((12873./20) + t*((-2373./2) + t*((4557./4)
              + t*((-9583./16) + t*((6543./40) + t*(-4361./240))))))) * h_1;
    t--;
    yphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    dyphi[3] = (t*((-49./18) + t*t*((-959./36) + t*((12845./144)
              + t*((-727./8) + t*(4361./144)))))) * h_1;
    t--;
    yphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    dyphi[4] = (t*((-49./18) + t*t*((-959./36) + t*((-12845./144)
              + t*((-727./8) + t*(-4361./144)))))) * h_1;
    t--;
    yphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    dyphi[5] = ((8617./60) + t*((12873./20) + t*((2373./2) + t*((4557./4)
              + t*((9583./16) + t*((6543./40) + t*(4361./240))))))) * h_1;
    t--;
    yphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    dyphi[6] = ((-25949./20) + t*((-117131./36) + t*((-6741./2)
            + t*((-66437./36) + t*((-81109./144) + t*((-727./8)
                  + t*(-4361./720))))))) * h_1;
    t--;
    yphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));
    dyphi[7] = ((7456./5) + t*((117572./45) + t*(1899 + t*((26383./36)
              + t*((22807./144) + t*((727./40) + t*(623./720))))))) * h_1;

    /* find t for z dimension, compute zphi and dzphi */
    t = dz_h - (double) klo;
    zphi[0] = (3632./5) + t*((-7456./5) + t*((58786./45) + t*(-633
            + t*((26383./144) + t*((-22807./720) + t*((727./240)
                  + t*(-89./720)))))));
    dzphi[0] = ((-7456./5) + t*((117572./45) + t*(-1899 + t*((26383./36)
              + t*((-22807./144) + t*((727./40) + t*(-623./720))))))) * h_1;
    t--;
    zphi[1] = -440 + t*((25949./20) + t*((-117131./72) + t*((2247./2)
            + t*((-66437./144) + t*((81109./720) + t*((-727./48)
                  + t*(623./720)))))));
    dzphi[1] = ((25949./20) + t*((-117131./36) + t*((6741./2)
            + t*((-66437./36) + t*((81109./144) + t*((-727./8)
                  + t*(4361./720))))))) * h_1;
    t--;
    zphi[2] = (138./5) + t*((-8617./60) + t*((12873./40) + t*((-791./2)
            + t*((4557./16) + t*((-9583./80) + t*((2181./80)
                  + t*(-623./240)))))));
    dzphi[2] = ((-8617./60) + t*((12873./20) + t*((-2373./2) + t*((4557./4)
              + t*((-9583./16) + t*((6543./40) + t*(-4361./240))))))) * h_1;
    t--;
    zphi[3] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((2569./144)
            + t*((-727./48) + t*(623./144)))));
    dzphi[3] = (t*((-49./18) + t*t*((-959./36) + t*((12845./144)
              + t*((-727./8) + t*(4361./144)))))) * h_1;
    t--;
    zphi[4] = 1 + t*t*((-49./36) + t*t*((-959./144) + t*((-2569./144)
            + t*((-727./48) + t*(-623./144)))));
    dzphi[4] = (t*((-49./18) + t*t*((-959./36) + t*((-12845./144)
              + t*((-727./8) + t*(-4361./144)))))) * h_1;
    t--;
    zphi[5] = (138./5) + t*((8617./60) + t*((12873./40) + t*((791./2)
            + t*((4557./16) + t*((9583./80) + t*((2181./80)
                  + t*(623./240)))))));
    dzphi[5] = ((8617./60) + t*((12873./20) + t*((2373./2) + t*((4557./4)
              + t*((9583./16) + t*((6543./40) + t*(4361./240))))))) * h_1;
    t--;
    zphi[6] = -440 + t*((-25949./20) + t*((-117131./72) + t*((-2247./2)
            + t*((-66437./144) + t*((-81109./720) + t*((-727./48)
                  + t*(-623./720)))))));
    dzphi[6] = ((-25949./20) + t*((-117131./36) + t*((-6741./2)
            + t*((-66437./36) + t*((-81109./144) + t*((-727./8)
                  + t*(-4361./720))))))) * h_1;
    t--;
    zphi[7] = (3632./5) + t*((7456./5) + t*((58786./45) + t*(633
            + t*((26383./144) + t*((22807./720) + t*((727./240)
                  + t*(89./720)))))));
    dzphi[7] = ((7456./5) + t*((117572./45) + t*(1899 + t*((26383./36)
              + t*((22807./144) + t*((727./40) + t*(623./720))))))) * h_1;

    /* determine 64=4*4*4 eh grid stencil contribution to force */
    fv.x = fv.y = fv.z = 0;
    for (k = 0;  k < 8;  k++) {
      koff = (k + klo) * ndim;
      for (j = 0;  j < 8;  j++) {
        jkoff = (koff + (j + jlo)) * ndim;
        c.x = yphi[j] * zphi[k];
        c.y = dyphi[j] * zphi[k];
        c.z = yphi[j] * dzphi[k];
        for (i = 0;  i < 8;  i++) {
          index = jkoff + (i + ilo);
          ASSERT(&eh[index]
              == mgrid_lattice_elem(&(mg->egrid[0]), i+ilo, j+jlo, k+klo));
          fv.x += eh[index] * dxphi[i] * c.x;
          fv.y += eh[index] * xphi[i] * c.y;
          fv.z += eh[index] * xphi[i] * c.z;
        }
      }
    }

    /* update force */
    f[n].x -= q[n] * fv.x;
    f[n].y -= q[n] * fv.y;
    f[n].z -= q[n] * fv.z;

  } /* end loop over n */

  /* compute potential */
  u = 0.0;
  for (n = 0;  n < nelems;  n++) {
    u += qvh[n] * evh[n];
  }

  /* must subtract self potential, save long range potential, add to total */
  sys->u_long = (0.5 * u) - u_self;
  sys->u_elec += sys->u_long;

  /* if separate long range force array, add to total force */
  if (sys->f_long != NULL) {
    for (n = 0;  n < natoms;  n++) {
      sys->f_elec[n].x += sys->f_long[n].x;
      sys->f_elec[n].y += sys->f_long[n].y;
      sys->f_elec[n].z += sys->f_long[n].z;
    }
  }

  return 0;
}



/*
 * constants for grid transfer operations
 * cubic numerical Hermite interpolant
 */

/* length of stencil */
enum { nstencil = 9 };

/* phi interpolating function along one dimension of grid stencil */
static const double phi[nstencil] = { -5./2048, 49./2048, -245./2048,
  1225./2048, 1, 1225./2048, -245./2048, 49./2048, -5./2048 };

/* stencil offsets from a central grid point on a finer grid level */
/* (these offsets are where phi weights above have been evaluated) */
static const int32 offset[nstencil] = { -7, -5, -3, -1, 0, 1, 3, 5, 7 };


int restrict_grid(Mgrid *mg, int32 level)
{
  double q2h_sum;
  double cjk;

  /* lattices of charge */
  const double *qh = (double *)(mg->qgrid[level].data);  /* finer grid */
  double *q2h = (double *)(mg->qgrid[level+1].data);     /* coarser grid */

  /* finer grid - grid size and labeling identical in each dimension */
  const int32 na1 = mg->qgrid[level].ia;      /* lowest grid index */
  const int32 nb1 = mg->qgrid[level].ib;      /* highest grid index */
  const int32 ndim1 = mg->qgrid[level].ni;    /* number points along dim */

  /* coarser grid - grid size and labeling identical in each dimension */
  const int32 na2 = mg->qgrid[level+1].ia;    /* lowest grid index */
  const int32 nb2 = mg->qgrid[level+1].ib;    /* highest grid index */
  const int32 ndim2 = mg->qgrid[level+1].ni;  /* number points along dim */

  /* other variables */
  int32 i1, j1, k1, index1, jk1off, k1off;
  int32 i2, j2, k2, index2, jk2off, k2off;
  int32 i, j, k;

  /* loop over coarser grid points */
  for (k2 = na2;  k2 <= nb2;  k2++) {
    k2off = k2 * ndim2;  /* coarser grid index offset for k-coord */
    k1 = k2 * 2;         /* k-coord of same-space point on finer grid */
    for (j2 = na2;  j2 <= nb2;  j2++) {
      jk2off = (k2off + j2) * ndim2;  /* add offset for j-coord coarser */
      j1 = j2 * 2;                    /* j-coord same-space finer grid */
      for (i2 = na2;  i2 <= nb2;  i2++) {
        index2 = jk2off + i2;         /* index in coarser grid */
        i1 = i2 * 2;                  /* i-coord same-space finer grid */

        /* sum weighted charge contribution from finer grid stencil */
        q2h_sum = 0.0;
        for (k = 0;  k < nstencil;  k++) {
          /* early loop termination if outside lattice */
          if (k1 + offset[k] < na1) continue;
          else if (k1 + offset[k] > nb1) break;
          k1off = (k1 + offset[k]) * ndim1;  /* offset k-coord finer grid */
          for (j = 0;  j < nstencil;  j++) {
            /* early loop termination if outside lattice */
            if (j1 + offset[j] < na1) continue;
            else if (j1 + offset[j] > nb1) break;
            jk1off = (k1off + (j1 + offset[j])) * ndim1;  /* add offset j */
            cjk = phi[j] * phi[k];              /* mult weights in each dim */
            for (i = 0;  i < nstencil;  i++) {
              /* early loop termination if outside lattice */
              if (i1 + offset[i] < na1) continue;
              else if (i1 + offset[i] > nb1) break;
              index1 = jk1off + (i1 + offset[i]);    /* index in finer grid */
              ASSERT(&qh[index1]
                  == mgrid_lattice_elem(&(mg->qgrid[level]),
                    i1 + offset[i], j1 + offset[j], k1 + offset[k]));
              q2h_sum += qh[index1] * phi[i] * cjk;  /* sum weighted charge */
            }
          }
        } /* end loop over finer grid stencil */

        ASSERT(&q2h[index2]
            == mgrid_lattice_elem(&(mg->qgrid[level+1]), i2, j2, k2));
        q2h[index2] = q2h_sum;  /* store charge to coarser grid */

      }
    }
  } /* end loop over each coarser grid point */

  return 0;
}


int prolong_grid(Mgrid *mg, int32 level)
{
  double cjk;

  /* lattices of potential */
  double *eh = (double *)(mg->egrid[level].data);           /* finer grid */
  const double *e2h = (double *)(mg->egrid[level+1].data);  /* coarser grid */

  /* finer grid - grid size and labeling identical in each dimension */
  const int32 na1 = mg->egrid[level].ia;      /* lowest grid index */
  const int32 nb1 = mg->egrid[level].ib;      /* highest grid index */
  const int32 ndim1 = mg->egrid[level].ni;    /* number points along dim */

  /* coarser grid - grid size and labeling identical in each dimension */
  const int32 na2 = mg->egrid[level+1].ia;    /* lowest grid index */
  const int32 nb2 = mg->egrid[level+1].ib;    /* highest grid index */
  const int32 ndim2 = mg->egrid[level+1].ni;  /* number points along dim */

  /* other variables */
  int32 i1, j1, k1, index1, jk1off, k1off;
  int32 i2, j2, k2, index2, jk2off, k2off;
  int32 i, j, k;

  /* loop over coarser grid points */
  for (k2 = na2;  k2 <= nb2;  k2++) {
    k2off = k2 * ndim2;  /* coarser grid index offset for k-coord */
    k1 = k2 * 2;         /* k-coord of same-space point on finer grid */
    for (j2 = na2;  j2 <= nb2;  j2++) {
      jk2off = (k2off + j2) * ndim2;  /* add offset for j-coord coarser */
      j1 = j2 * 2;                    /* j-coord same-space finer grid */
      for (i2 = na2;  i2 <= nb2;  i2++) {
        index2 = jk2off + i2;         /* index in coarser grid */
        i1 = i2 * 2;                  /* i-coord same-space finer grid */

        /* sum weighted potential contribution to finer grid stencil */
        for (k = 0;  k < nstencil;  k++) {
          /* early loop termination if outside lattice */
          if (k1 + offset[k] < na1) continue;
          else if (k1 + offset[k] > nb1) break;
          k1off = (k1 + offset[k]) * ndim1;  /* offset k-coord finer grid */
          for (j = 0;  j < nstencil;  j++) {
            /* early loop termination if outside lattice */
            if (j1 + offset[j] < na1) continue;
            else if (j1 + offset[j] > nb1) break;
            jk1off = (k1off + (j1 + offset[j])) * ndim1;  /* add offset j */
            cjk = phi[j] * phi[k];              /* mult weights in each dim */
            for (i = 0;  i < nstencil;  i++) {
              /* early loop termination if outside lattice */
              if (i1 + offset[i] < na1) continue;
              else if (i1 + offset[i] > nb1) break;
              index1 = jk1off + (i1 + offset[i]);    /* index in finer grid */
              ASSERT(&eh[index1]
                  == mgrid_lattice_elem(&(mg->egrid[level]),
                    i1 + offset[i], j1 + offset[j], k1 + offset[k]));
              ASSERT(&e2h[index2]
                  == mgrid_lattice_elem(&(mg->egrid[level+1]), i2, j2, k2));
              /* sum weighted potential from coarser grid */
              eh[index1] += e2h[index2] * phi[i] * cjk;
            }
          }
        } /* end loop over finer grid stencil */

      }
    }
  } /* end loop over each coarser grid point */

  return 0;
}


int direct_sum(Mgrid *mg, int32 level)
{
  double eh_sum;
  double scaling = mg->scaling[level];     /* scaling factor for gd */

  /* lattices of charge and potential */
  const double *qh = (double *)(mg->qgrid[level].data);
  double *eh = (double *)(mg->egrid[level].data);
  const int32 ndim = mg->qgrid[level].ni;  /* number points in each dim */
  const int32 na = mg->qgrid[level].ia;    /* lowest numbered point */
  const int32 nb = mg->qgrid[level].ib;    /* highest numbered point */

  /*
   * for last level use different block of weights that
   * produces pairwise interactions between all grid points
   */

  /* lattice of weights for short range charge influence on potential */
  const double *gd = (level == mg->param.nlevels - 1 ?
      (double *)(mg->glast.data) : (double *)(mg->gdsum.data));

  /* dimension of gd lattice */
  const int32 ngdim = (level == mg->param.nlevels - 1 ?
      mg->glast.ni : mg->gdsum.ni);

  /* "radius" of gd "sphere" */
  const int32 ngrad = (level == mg->param.nlevels - 1 ?
      mg->glast_radius : mg->gdsum_radius);

  int32 i, j, k, ig, jg, kg;
  int32 iga, igb, jga, jgb, kga, kgb;
  int32 index, jkoff, koff;
  int32 nindex, jknoff, knoff;
  int32 ngindex, jkgoff, kgoff;

  PTR(mg->glast.data);
  PTR(gd);

  INT(mg->glast.ni);
  INT(ngdim);

  INT(mg->glast_radius);
  INT(ngrad);

#ifdef DEBUG_WATCH
  mgrid_lattice_print(&(mg->glast));
#endif

  /* loop over all grid points */
  for (k = na;  k <= nb;  k++) {
    /* find k-coord index ranges that keep gd on grid */
    kga = (na - k > -ngrad ? na - k : -ngrad);
    kgb = (nb - k < ngrad ? nb - k : ngrad);

    koff = k * ndim;  /* grid index offset for k-coord */

    for (j = na;  j <= nb;  j++) {
      /* find j-coord index ranges that keep gd on grid */
      jga = (na - j > -ngrad ? na - j : -ngrad);
      jgb = (nb - j < ngrad ? nb - j : ngrad);

      jkoff = (koff + j) * ndim;  /* add offset for j-coord */

      for (i = na;  i <= nb;  i++) {
        /* find i-coord index ranges that keep gd on grid */
        iga = (na - i > -ngrad ? na - i : -ngrad);
        igb = (nb - i < ngrad ? nb - i : ngrad);

        index = jkoff + i;  /* index for potential grid point */

#ifdef DEBUG_WATCH
        printf("(%d, %d, %d):  sphere index (%d..%d, %d..%d, %d..%d)\n",
            i, j, k, iga, igb, jga, jgb, kga, kgb);
#endif
        /* sum over "sphere" of weighted charge */
        eh_sum = 0.0;
        for (kg = kga;  kg <= kgb;  kg++) {
          knoff = (k + kg) * ndim;  /* grid index offset for k-coord */
          kgoff = kg * ngdim;       /* gd index offset for k-coord */

          for (jg = jga;  jg <= jgb;  jg++) {
            jknoff = (knoff + (j + jg)) * ndim;  /* add offset for j-coord */
            jkgoff = (kgoff + jg) * ngdim;       /* gd add offset j-coord */

            for (ig = iga;  ig <= igb;  ig++) {
              nindex = jknoff + (i + ig);  /* index of charge contribution */
              ngindex = jkgoff + ig;       /* index of gd weighting factor */
              ASSERT(&qh[nindex]
                  == mgrid_lattice_elem(&(mg->qgrid[level]),
                    i + ig, j + jg, k + kg));
              ASSERT(&gd[ngindex]
                  == mgrid_lattice_elem((level == mg->param.nlevels - 1 ?
                      &(mg->glast) : &(mg->gdsum)), ig, jg, kg));
              eh_sum += qh[nindex] * gd[ngindex];  /* sum weighted charge */
            }
          }
        } /* end loop over "sphere" of charge */

        ASSERT(&eh[index]
            == mgrid_lattice_elem(&(mg->egrid[level]), i, j, k));
        eh[index] = scaling * eh_sum;  /* scale and store potential */
      }
    }
  } /* end loop over all grid points */

  INT(level);
#ifdef DEBUG_WATCH
  mgrid_lattice_print(&(mg->egrid[level]));
#endif

  return 0;
}
